
import os

__all__ = ['demo_data', 'demo_campbell_scientific_data', 'demo_merra2_NW',
           'demo_merra2_NE', 'demo_merra2_SE', 'demo_merra2_SW', 'demo_windographer_data']

demo_campbell_scientific_data = os.path.join(os.path.dirname(__file__), 'demo', 'campbell_scientific_demo_data.csv')
demo_data = os.path.join(os.path.dirname(__file__), 'demo', 'demo_data.csv')
demo_windographer_data = os.path.join(os.path.dirname(__file__), 'demo', 'windographer_demo_data.txt')

demo_merra2_NW = os.path.join(os.path.dirname(__file__), 'demo', 'MERRA-2_NW_2000-01-01_2017-06-30.csv')
demo_merra2_NE = os.path.join(os.path.dirname(__file__), 'demo', 'MERRA-2_NE_2000-01-01_2017-06-30.csv')
demo_merra2_SE = os.path.join(os.path.dirname(__file__), 'demo', 'MERRA-2_SE_2000-01-01_2017-06-30.csv')
demo_merra2_SW = os.path.join(os.path.dirname(__file__), 'demo', 'MERRA-2_SW_2000-01-01_2017-06-30.csv')


def datasets_available():
    """
    Example datasets that can be used with the library.


    **Example usage**
    ::
        import brightwind as bw

        all_datasets_available = ['demo_data', 'demo_campbell_scientific_data', 'demo_merra2_NW',
           'demo_merra2_NE', 'demo_merra2_SE', 'demo_merra2_SW', 'demo_windographer_data']
        demo_data = bw.load_campbell_scientific(bw.datasets.demo_campbell_scientific_data)
        demo_data = bw.load_csv(bw.datasets.demo_data)
        demo_windog_data = bw.load_windographer_txt(bw.datasets.demo_windographer_data)

    """

    return None



#     brightwind is a library that provides wind analysts with easy to use tools for working with meteorological data.
#     Copyright (C) 2018 Stephen Holleran, Inder Preet
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Lesser General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pandas as pd
import numpy as np
import datetime
import calendar
from math import e
from brightwind.analyse import plot as plt
# noinspection PyProtectedMember
from brightwind.analyse.analyse import dist_by_dir_sector, dist_12x24, coverage, _convert_df_to_series
from ipywidgets import FloatProgress
from IPython.display import display
from IPython.display import clear_output
import re
import warnings

pd.options.mode.chained_assignment = None

__all__ = ['Shear']


class Shear:

    class TimeSeries:

        def __init__(self, wspds, heights, min_speed=3, calc_method='power_law', max_plot_height=None,
                     maximise_data=False):
            """
            Calculates alpha, using the power law, or the roughness coefficient, using the log law, for each timestamp
            of a wind series.

           :param wspds: pandas DataFrame, list of pandas.Series or list of wind speeds to be used for calculating shear.
           :type wspds:  pandas.DataFrame, list of pandas.Series or list.
           :param heights: List of anemometer heights.
           :type heights: list
           :param min_speed: Only speeds higher than this would be considered for calculating shear, default is 3.
           :type min_speed: float
           :param calc_method: method to use for calculation, either 'power_law' (returns alpha) or 'log_law'
                              (returns the roughness coefficient).
           :type calc_method: str
           :param max_plot_height: height to which the wind profile plot is extended.
           :type max_plot_height: float
           :param maximise_data: If maximise_data is True, calculations will be carried out on all data where two or
                                 more anemometers readings exist for a timestamp. If False, calculations will only be
                                 carried out on timestamps where readings exist for all anemometers.
           :type maximise_data: Boolean
           :return TimeSeries object containing calculated alpha/roughness coefficient values, a plot
                   and other data.
           :rtype TimeSeries object

           **Example usage**
           ::
                import brightwind as bw
                import pprint

                # Load anemometer data to calculate exponents
                data = bw.load_csv(C:\\Users\\Stephen\\Documents\\Analysis\\demo_data)
                anemometers = data[['Spd80mS', 'Spd60mS','Spd40mS']]
                heights = [80, 60, 40]

                # Using with a DataFrame of wind speeds
                timeseries_power_law = bw.Shear.TimeSeries(anemometers, heights, maximise_data=True)
                timeseries_log_law = bw.Shear.TimeSeries(anemometers, heights, calc_method='log_law',
                                                         max_plot_height=120)

                # Get the alpha or roughness values calculated
                timeseries_power_law.alpha
                timeseries_log_law.roughness

                # View plot
                timeseries_power_law.plot
                timeseries_log_law.plot

                # View input anemometer data
                timeseries_power_law.wspds
                timeseries_log_law.wspds

                # View other information
                pprint.pprint(timeseries_power_law.info)
                pprint.pprint(timeseries_log_law.info)

           """
            print('This may take a while...')

            wspds, cvg = Shear._data_prep(wspds=wspds, heights=heights, min_speed=min_speed, maximise_data=maximise_data)

            if calc_method == 'power_law':
                alpha_c = (wspds[(wspds > min_speed).all(axis=1)].apply(Shear._calc_power_law, heights=heights,
                                                                        return_coeff=True,
                                                                        maximise_data=maximise_data, axis=1))
                alpha = pd.Series(alpha_c.iloc[:, 0], name='alpha')
                self._alpha = alpha

            elif calc_method == 'log_law':
                slope_intercept = (wspds[(wspds > min_speed).all(axis=1)].apply(Shear._calc_log_law, heights=heights,
                                                                                return_coeff=True,
                                                                                maximise_data=maximise_data, axis=1))
                slope = slope_intercept.iloc[:, 0]
                intercept = slope_intercept.iloc[:, 1]
                roughness_coefficient = pd.Series(Shear._calc_roughness(slope=slope, intercept=intercept),
                                                  name='roughness_coefficient')
                self._roughness = roughness_coefficient

            clear_output()
            avg_plot = Shear.Average(wspds=wspds, heights=heights, calc_method=calc_method,
                                     max_plot_height=max_plot_height)

            self.origin = 'TimeSeries'
            self.calc_method = calc_method
            self.wspds = wspds
            self.plot = avg_plot.plot
            self.info = Shear._create_info(self, heights=heights, cvg=cvg, min_speed=min_speed)

        @property
        def alpha(self):
            return self._alpha

        @property
        def roughness(self):
            return self._roughness

        def apply(self, wspds, height, shear_to):
            """"
            Applies shear calculated to a wind speed time series and scales wind speed from one height to
            another for each matching timestamp.

           :param self: TimeSeries object to use when applying shear to the data.
           :type self: TimeSeries object
           :param wspds: Wind speed time series to apply shear to.
           :type wspds: pandas.Series
           :param height: height of above wspds.
           :type height: float
           :param shear_to: height to which wspds should be scaled to.
           :type shear_to: float
           :return: a pandas.Series of the scaled wind speeds.
           :rtype: pandas.Series

            **Example Usage**
            ::
                import brightwind as bw

                # Load anemometer data to calculate exponents
                data = bw.load_csv(C:\\Users\\Stephen\\Documents\\Analysis\\demo_data)
                anemometers = data[['Spd80mS', 'Spd60mS','Spd40mS']]
                heights = [80, 60, 40]

                # Get power law object
                timeseries_power_law = bw.Shear.TimeSeries(anemometers, heights)
                timeseries_log_law = bw.Shear.TimeSeries(anemometers, heights, calc_method='log_law')

                # Scale wind speeds using calculated exponents
                timeseries_power_law.apply(data['Spd40mN'], height=40, shear_to=70)
                timeseries_log_law.apply(data['Spd40mN'], height=40, shear_to=70)
               """

            return Shear._apply(self, wspds, height, shear_to)

    class TimeOfDay:

        def __init__(self, wspds, heights, min_speed=3, calc_method='power_law', by_month=True, segment_start_time=7,
                     segments_per_day=24, plot_type='line'):
            """
            Calculates alpha, using the power law, or the roughness coefficient, using the log law, for a wind series
            binned by time of the day and (optionally by) month, depending on the user's inputs. The alpha/roughness
            coefficient values are calculated based on the average wind speeds at each measurement height in each bin.

            :param wspds: pandas.DataFrame, list of pandas.Series or list of wind speeds to be used for calculating
             shear.
            :type wspds:  pandas.DataFrame, list of pandas.Series or list.
            :param heights: List of anemometer heights..
            :type heights: list
            :param min_speed: Only speeds higher than this would be considered for calculating shear, default is 3
            :type min_speed: float
            :param calc_method: method to use for calculation, either 'power_law' (returns alpha) or 'log_law'
                                (returns the roughness coefficient).
            :type calc_method: str
            :param by_month: If True, calculate alpha or roughness coefficient values for each daily segment and month.
                             If False, average alpha or roughness coefficient values are calculated for each daily
                             segment across all months.
            :type by_month: Boolean
            :param segment_start_time: Starting time for first segment.
            :type segment_start_time: int
            :param segments_per_day: Number of segments into which each 24 period is split. Must be a divisor of 24.
            :type segments_per_day: int
            :param plot_type: Type of plot to be generated. Options include 'line', 'step' and '12x24'.
            :type plot_type: str
            :return: TimeOfDay object containing calculated alpha/roughness coefficient values, a plot
                     and other data.
            :rtype: TimeOfDay object

            **Example usage**
            ::
                import brightwind as bw
                import pprint

                # Load anemometer data to calculate exponents
                data = bw.load_csv(C:\\Users\\Stephen\\Documents\\Analysis\\demo_data)
                anemometers = data[['Spd80mS', 'Spd60mS','Spd40mS']]
                heights = [80, 60, 40]

                # Using with a DataFrame of wind speeds
                timeofday_power_law = bw.Shear.TimeOfDay(anemometers, heights, daily_segments=2, segment_start_time=7)
                timeofday_log_law = bw.Shear.TimeOfDay(anemometers, heights, calc_method='log_law', by_month=False)

                # Get alpha or roughness values calculated
                timeofday_power_law.alpha
                timeofday_log_law.roughness

                # View plot
                timeofday_power_law.plot
                timeofday_log_law.plot

                # View input data
                timeofday_power_law.wspds
                timeofday_log_law.wspds

                # View other information
                pprint.pprint(timeofday_power_law.info)
                pprint.pprint(timeofday_log_law.info)

            """

            wspds, cvg = Shear._data_prep(wspds=wspds, heights=heights, min_speed=min_speed)

            # initialise empty series for later use
            start_times = pd.Series([])
            time_wspds = pd.Series([])
            mean_time_wspds = pd.Series([])
            c = pd.Series([])
            slope = pd.Series([])
            intercept = pd.Series([])
            alpha = pd.Series([])
            roughness = pd.Series([])
            slope_df = pd.DataFrame([])
            intercept_df = pd.DataFrame([])
            roughness_df = pd.DataFrame([])
            alpha_df = pd.DataFrame([])

            # time of day shear calculations
            interval = int(24 / segments_per_day)

            if by_month is False and plot_type == '12x24':
                raise ValueError("12x24 plot is only possible when 'by_month=True'")

            if not int(segment_start_time) % 1 == 0:
                raise ValueError("'segment_start_time' must be an integer between 0 and 24'")

            if not (24 % segments_per_day == 0) | (segments_per_day == 1):
                raise ValueError("'segments_per_day' must be a divisor of 24'")

            segment_start_time = str(segment_start_time)
            start_times[0] = datetime.datetime.strptime(segment_start_time, '%H')
            dt = datetime.timedelta(hours=interval)

            # extract wind speeds for each daily segment
            for i in range(1, segments_per_day):
                start_times[i] = start_times[i - 1] + dt

            # extract wind speeds for each month
            for j in range(0, 12):

                anemometers_df = wspds[wspds.index.month == j + 1]
                for i in range(0, segments_per_day):

                    if segments_per_day == 1:
                        mean_time_wspds[i] = anemometers_df.mean().dropna()

                    elif i == segments_per_day - 1:
                        start_times[i] = start_times[i].strftime("%H:%M:%S")
                        start = str(start_times[i].time())
                        end = str(start_times[0].time())
                        time_wspds[i] = pd.DataFrame(anemometers_df).between_time(start, end, include_end=False)
                        mean_time_wspds[i] = time_wspds[i][(time_wspds[i] > min_speed).all(axis=1)].mean().dropna()

                    else:
                        start_times[i] = start_times[i].strftime("%H:%M:%S")
                        start = str(start_times[i].time())
                        end = str(start_times[i + 1].time())
                        time_wspds[i] = pd.DataFrame(anemometers_df).between_time(start, end, include_end=False)
                        mean_time_wspds[i] = time_wspds[i][(time_wspds[i] > min_speed).all(axis=1)].mean().dropna()

                # calculate shear
                if calc_method == 'power_law':
                    for i in range(0, len(mean_time_wspds)):
                        alpha[i], c[i] = Shear._calc_power_law(mean_time_wspds[i].values, heights, return_coeff=True)
                    alpha_df = pd.concat([alpha_df, alpha], axis=1)

                if calc_method == 'log_law':
                    for i in range(0, len(mean_time_wspds)):
                        slope[i], intercept[i] = Shear._calc_log_law(mean_time_wspds[i].values, heights,
                                                                     return_coeff=True)
                        roughness[i] = Shear._calc_roughness(slope=slope[i], intercept=intercept[i])
                    roughness_df = pd.concat([roughness_df, roughness], axis=1)
                    slope_df = pd.concat([slope_df, slope], axis=1)
                    intercept_df = pd.concat([intercept_df, intercept], axis=1)

            # error check
            if mean_time_wspds.shape[0] == 0:
                raise ValueError('None of the input wind speeds are greater than the min_speed, cannot calculate shear')

            if calc_method == 'power_law':
                alpha_df.index = start_times
                alpha_df.index = alpha_df.index.time
                alpha_df.sort_index(inplace=True)

                if by_month is True:
                    alpha_df.columns = calendar.month_abbr[1:13]
                    self.plot = plt.plot_shear_time_of_day(Shear._fill_df_12x24(alpha_df), calc_method=calc_method,
                                                           plot_type=plot_type)

                else:
                    alpha_df = pd.DataFrame(alpha_df.mean(axis=1))
                    alpha_df.columns = ['12 Month Average']
                    self.plot = plt.plot_shear_time_of_day(pd.DataFrame((Shear._fill_df_12x24(alpha_df)).iloc[:, 0]),
                                                           calc_method=calc_method, plot_type=plot_type)

                alpha_df.index.name = 'segment_start_time'
                self._alpha = alpha_df

            if calc_method == 'log_law':
                roughness_df.index = slope_df.index = intercept_df.index = start_times
                roughness_df.index = slope_df.index = intercept_df.index = roughness_df.index.time
                roughness_df.sort_index(inplace=True)

                slope_df.sort_index(inplace=True)
                intercept_df.sort_index(inplace=True)

                if by_month is True:
                    roughness_df.columns = slope_df.columns = intercept_df.columns = calendar.month_abbr[1:13]
                    self.plot = plt.plot_shear_time_of_day(Shear._fill_df_12x24(roughness_df),
                                                           calc_method=calc_method, plot_type=plot_type)
                else:
                    slope_df = pd.DataFrame(slope_df.mean(axis=1))
                    intercept_df = pd.DataFrame(intercept_df.mean(axis=1))
                    roughness_df = pd.DataFrame(roughness_df.mean(axis=1))
                    roughness_df.columns = slope_df.columns = intercept_df.columns = ['12_month_average']
                    self.plot = plt.plot_shear_time_of_day(
                        pd.DataFrame(Shear._fill_df_12x24(roughness_df).iloc[:, 0]),
                        calc_method=calc_method, plot_type=plot_type)

                roughness_df.index.name = 'segment_start_time'
                self._roughness = roughness_df

            self.calc_method = calc_method
            self.wspds = wspds
            self.origin = 'TimeOfDay'
            self.info = Shear._create_info(self, heights=heights, cvg=cvg, min_speed=min_speed,
                                           segment_start_time=segment_start_time, segments_per_day=segments_per_day)

        @property
        def alpha(self):
            return self._alpha

        @property
        def roughness(self):
            return self._roughness

        def apply(self, wspds, height, shear_to):
            """
            Applies shear calculated to a wind speed time series by time of day (and optionally by month) to scale
            wind speed from one height to another.

           :param self: TimeOfDay object to use when applying shear to the data.
           :type self: TimeOfDay object
           :param wspds: Wind speed time series to apply shear to.
           :type wspds: pandas.Series
           :param height: height of above wspds.
           :type height: float
           :param shear_to: height to which wspds should be scaled.
           :type shear_to: float
           :return: a pandas.Series of the scaled wind speeds.
           :rtype: pandas.Series

            **Example Usage**
            ::
                import brightwind as bw

                # Load anemometer data to calculate exponents
                data = bw.load_csv(C:\\Users\\Stephen\\Documents\\Analysis\\demo_data)
                anemometers = data[['Spd80mS', 'Spd60mS','Spd40mS']]
                heights = [80, 60, 40]

                # Get power law object
                timeofday_power_law = bw.Shear.TimeOfDay(anemometers, heights)
                timeofday_log_law = bw.Shear.TimeOfDay(anemometers, heights, calc_method='log_law')

                # Scale wind speeds using calculated exponents
                timeofday_power_law.apply(data['Spd40mN'], height=40, shear_to=70)
                timeofday_log_law.apply(data['Spd40mN'], height=40, shear_to=70)

            """

            return Shear._apply(self, wspds, height, shear_to)

    class Average:

        def __init__(self, wspds, heights, min_speed=3, calc_method='power_law', plot_both=False, max_plot_height=None):
            """
             Calculates alpha, using the power law, or the roughness coefficient, using the log law, based on the
             average wind speeds of each supplied time series.

            :param wspds: pandas.DataFrame, list of pandas.Series or list of wind speeds to be used for calculating
             shear.
            :type wspds:  pandas.DataFrame, list of pandas.Series or list.
            :param heights: List of anemometer heights
            :type heights: list
            :param min_speed: Only speeds higher than this would be considered for calculating shear, default is 3.
            :type min_speed: float
            :param calc_method: method to use for calculation, either 'power_law' (returns alpha) or 'log_law'
                                      (returns the roughness coefficient).
            :type calc_method: str
            :param max_plot_height: height to which the wind profile plot is extended.
            :type max_plot_height: float
            :return:  Average object containing calculated alpha/roughness coefficient values, a plot and other data.
            :rtype: Average object

            **Example usage**
            ::
                import brightwind as bw
                import pprint

                # Load anemometer data to calculate exponents
                data = bw.load_csv(C:\\Users\\Stephen\\Documents\\Analysis\\demo_data)
                anemometers = data[['Spd80mS', 'Spd60mS','Spd40mS']]
                heights = [80, 60, 40]

                # Using with a DataFrame of wind speeds
                average_power_law = bw.Shear.Average(anemometers, heights)
                average_log_law = bw.Shear.Average(anemometers, heights, calc_method='log_law', max_plot_height=120)

                # Get the alpha or roughness values calculated
                average_power_law.alpha
                average_log_law.roughness

                # View plot
                average_power_law.plot
                average_log_law.plot

                # View input data
                average_power_law.wspds
                average_log_law.wspds

                # View other information
                pprint.pprint(average_power_law.info)
                pprint.pprint(average_log_law.info)

            """

            wspds, cvg = Shear._data_prep(wspds=wspds, heights=heights, min_speed=min_speed)

            mean_wspds = wspds.mean(axis=0)

            if mean_wspds.shape[0] == 0:
                raise ValueError('None of the input wind speeds are greater than the min_speed, cannot calculate shear')

            if calc_method == 'power_law':
                alpha, c = Shear._calc_power_law(mean_wspds.values, heights, return_coeff=True)
                if plot_both is True:
                    slope, intercept = Shear._calc_log_law(mean_wspds.values, heights, return_coeff=True)
                    self.plot = plt.plot_power_law(plot_both=True, avg_alpha=alpha, avg_c=c, avg_slope=slope,
                                                   avg_intercept=intercept,
                                                   wspds=mean_wspds.values, heights=heights,
                                                   max_plot_height=max_plot_height)
                else:
                    self.plot = plt.plot_power_law(alpha, c, mean_wspds.values, heights,
                                                   max_plot_height=max_plot_height)
                self._alpha = alpha

            elif calc_method == 'log_law':
                slope, intercept = Shear._calc_log_law(mean_wspds.values, heights, return_coeff=True)
                roughness = Shear._calc_roughness(slope=slope, intercept=intercept)
                self._roughness = roughness
                if plot_both is True:
                    alpha, c = Shear._calc_power_law(mean_wspds.values, heights, return_coeff=True)
                    self.plot = plt.plot_power_law(avg_alpha=alpha, avg_c=c, avg_slope=slope, avg_intercept=intercept,
                                                   wspds=mean_wspds.values, heights=heights,
                                                   max_plot_height=max_plot_height)
                else:
                    self.plot = plt.plot_log_law(slope, intercept, mean_wspds.values, heights,
                                                 max_plot_height=max_plot_height)

            else:
                raise ValueError("Please enter a valid calculation method, either 'power_law' or 'log_law'.")

            self.wspds = wspds
            self.origin = 'Average'
            self.calc_method = calc_method
            self.info = Shear._create_info(self, heights=heights, min_speed=min_speed, cvg=cvg)

        @property
        def alpha(self):
            return self._alpha

        @property
        def roughness(self):
            return self._roughness

        def apply(self, wspds, height, shear_to):
            """
            Applies average shear calculated to a wind speed time series to scale wind speed from one height to another.

           :param self: Average object to use when applying shear to the data.
           :type self: Average object
           :param wspds: Wind speed time series to apply shear to.
           :type wspds: pandas.Series
           :param height: height of above wspds.
           :type height: float
           :param shear_to: height to which wspds should be scaled to.
           :type shear_to: float
           :return: a pandas.Series of the scaled wind speeds.
           :rtype: pandas.Series

            **Example Usage**
            ::
                import brightwind as bw

                # Load anemometer data to calculate exponents
                data = bw.load_csv(C:\\Users\\Stephen\\Documents\\Analysis\\demo_data)
                anemometers = data[['Spd80mS', 'Spd60mS','Spd40mS']]
                heights = [80, 60, 40]

                # Get power law object
                average_power_law = bw.Shear.Average(anemometers, heights)
                average_log_law = bw.Shear.Average(anemometers, heights, calc_method='log_law', max_plot_height=120)

               # Scale wind speeds using exponents
               average_power_law.apply(data['Spd40mN'], height=40, shear_to=70)
               average_log_law.apply(data['Spd40mN'], height=40, shear_to=70)

               """
            return Shear._apply(self, wspds, height, shear_to)

    class BySector:

        def __init__(self, wspds, heights, wdir, min_speed=3, calc_method='power_law', sectors=12,
                     direction_bin_array=None, direction_bin_labels=None):
            """
            Calculates alpha, using the power law, or the roughness coefficient, using the log law, for a wind series
            binned by direction. The alpha/roughness coefficient values are calculated based on the average wind speeds
            at each measurement height in each bin.


            :param wspds: pandas.DataFrame, list of pandas.Series or list of wind speeds to be used for calculating
            shear.
            :type wspds: pandas.DataFrame, list of pandas.Series or list.
            :param heights: List of anemometer heights
            :type heights: list
            :param wdir: Wind direction measurements
            :type wdir: pandas.DataFrame or Series
            :param: min_speed: Only speeds higher than this would be considered for calculating shear, default is 3.
            :type: min_speed: float
            :param calc_method: Method to use for calculation, either 'power_law' (returns alpha) or 'log_law'
                                (returns the roughness coefficient).
            :type calc_method: str
            :param sectors: Number of sectors for the shear to be calculated for.
            :type sectors: int
            :param direction_bin_array: Specific array of directional bins to be used. If None, bins are calculated
                                        by 360/sectors.
            :type direction_bin_array: array
            :param direction_bin_labels: Labels to be given to the above direction_bin array.
            :type direction_bin_labels: array
            :return: BySector object containing calculated alpha/roughness coefficient values, a plot and other data.
            :rtype: BySector object

             **Example usage**
            ::
                import brightwind as bw
                import pprint

                # Load anemometer data to calculate exponents
                data = bw.load_csv(C:\\Users\\Stephen\\Documents\\Analysis\\demo_data)
                anemometers = data[['Spd80mS', 'Spd60mS','Spd40mS']]
                heights = [80, 60, 40]
                directions = data['Dir78mS']

                # Calculate shear exponents using default bins ([345,15,45,75,105,135,165,195,225,255,285,315,345])
                by_sector_power_law= bw.Shear.BySector(anemometers, heights, directions)
                by_sector_log_law= bw.Shear.BySector(anemometers, heights, directions, calc_method='log_law')

                # Calculate shear exponents using custom bins
                custom_bins = [0,30,60,90,120,150,180,210,240,270,300,330,360]
                by_sector_power_law_custom_bins = bw.Shear.BySector(anemometers, heights, directions,
                direction_bin_array=custom_bins)

                # Get alpha or roughness values calculated
                by_sector_power_law.alpha
                by_sector_log_law.roughness

                # View plot
                by_sector_power_law.plot
                by_sector_log_law.plot

                # View input data
                by_sector_power_law.wspds
                by_sector_log_law.wspds

                # View other information
                pprint.pprint(by_sector_power_law.info)
                pprint.pprint(by_sector_log_law.info)

            """
            print('This may take a while...')
            wspds, cvg = Shear._data_prep(wspds=wspds, heights=heights, min_speed=min_speed)

            if direction_bin_array is not None:
                sectors = len(direction_bin_array) - 1

            wdir = _convert_df_to_series(wdir)
            mean_wspds = pd.Series([])
            mean_wspds_df = pd.DataFrame([])
            count_df = pd.DataFrame([])
            count = pd.Series([])

            for i in range(len(wspds.columns)):

                w = wspds.iloc[:, i]
                plot, mean_wspds[i] = dist_by_dir_sector(w, wdir, direction_bin_array=direction_bin_array,
                                                         aggregation_method='mean', return_data=True)

                plot, count[i] = dist_by_dir_sector(w, wdir, direction_bin_array=direction_bin_array,
                                                    aggregation_method='count', return_data=True)

                if i == 0:
                    mean_wspds_df = mean_wspds[i].copy()
                    count_df = count[i].copy()
                else:
                    mean_wspds_df = pd.concat([mean_wspds_df, mean_wspds[i]], axis=1)
                    count_df = pd.concat([count_df, count[i]], axis=1)

            count_df = count_df.mean(axis=1)
            wind_rose_plot, wind_rose_dist = dist_by_dir_sector(wspds.iloc[:, 0], wdir,
                                                                direction_bin_array=direction_bin_array,
                                                                direction_bin_labels=direction_bin_labels,
                                                                return_data=True)
            if calc_method == 'power_law':

                alpha = mean_wspds_df.apply(Shear._calc_power_law, heights=heights, return_coeff=False, axis=1)

                wind_rose_plot, wind_rose_dist = dist_by_dir_sector(wspds.iloc[:, 0], wdir,
                                                                    direction_bin_array=direction_bin_array,
                                                                    direction_bin_labels=direction_bin_labels,
                                                                    return_data=True)

                self.alpha_count = count_df
                self._alpha = pd.Series(alpha, name='alpha')
                clear_output()
                self.plot = plt.plot_shear_by_sector(scale_variable=alpha, wind_rose_data=wind_rose_dist,
                                                     calc_method=calc_method)

            elif calc_method == 'log_law':

                slope_intercept = mean_wspds_df.apply(Shear._calc_log_law, heights=heights, return_coeff=True, axis=1)

                slope = slope_intercept.iloc[:, 0]
                intercept = slope_intercept.iloc[:, 1]
                roughness = Shear._calc_roughness(slope=slope, intercept=intercept)
                self.roughness_count = count_df
                self._roughness = pd.Series(roughness, name='roughness_coefficient')
                clear_output()
                self.plot = plt.plot_shear_by_sector(scale_variable=roughness, wind_rose_data=wind_rose_dist,
                                                     calc_method=calc_method)

            else:
                raise ValueError("Please enter a valid calculation method, either 'power_law' or 'log_law'.")

            self.wspds = wspds
            self.wdir = wdir
            self.origin = 'BySector'
            self.sectors = sectors
            self.calc_method = calc_method
            self.info = Shear._create_info(self, heights=heights, cvg=cvg, min_speed=min_speed,
                                           direction_bin_array = direction_bin_array)

        @property
        def alpha(self):
            return self._alpha

        @property
        def roughness(self):
            return self._roughness

        def apply(self, wspds, wdir, height, shear_to):
            """
            Applies shear calculated to a wind speed time series by wind direction to scale
            wind speed from one height to another.

            :param self: BySector object to use when applying shear to the data
            :type self: BySector object
            :param wspds: Wind speed time series to apply shear to.
            :type wspds: pandas.Series
            :param wdir: Wind direction measurements of wspds, only required if shear is to be applied by direction
            sector.
            :type wdir: pandas.Series
            :param height: Height of wspds.
            :type height: float
            :param shear_to: Height to which wspds should be scaled to.
            :type shear_to: float
            :return: A pandas.Series of the scaled wind speeds.
            :rtype: pandas.Series

             **Example Usage**
             ::
                import brightwind as bw

                # Load anemometer data to calculate exponents
                data = bw.load_csv(C:\\Users\\Stephen\\Documents\\Analysis\\demo_data)
                anemometers = data[['Spd80mS', 'Spd60mS']]
                heights = [80, 60]
                directions = data[['Dir78mS']]

                # Calculate shear exponents
                by_sector_power_law = bw.Shear.BySector(anemometers, heights, directions)
                by_sector_log_law = bw.Shear.BySector(anemometers, heights, directions, calc_method='log_law')

                # Calculate shear exponents using default bins ([345,15,45,75,105,135,165,195,225,255,285,315,345])
                by_sector_power_law= bw.Shear.BySector(anemometers, heights, directions)

                # Scale wind speeds using exponents
                by_sector_power_law.apply(data['Spd40mN'], data['Dir38mS'], height=40, shear_to=70)
                by_sector_log_law.apply(data['Spd40mN'], data['Dir38mS'], height=40, shear_to=70)

            """

            return Shear._apply(self, wspds=wspds, height=height, shear_to=shear_to, wdir=wdir)

    @staticmethod
    def _log_roughness_scale(wspds, height, shear_to, roughness):
        """
        Scale wind speeds using the logarithmic wind shear law.

        :param wspds: wind speeds at height z1, U1
        :param height: z1
        :param shear_to: z2
        :param roughness: z0
        :return: Scaled wind speeds, U2
        :rtype: pandas.Series or float

        METHODOLOGY:

                                            U2 = (ln(z2/z0)/ln(z1/z0))U1
            Where:
                    - U2 is the wind speed at height z2
                    - U1 is the wind speed at height z1
                    - z1 is the lower height
                    - z2 is the upper height
                    - zo is the roughness coefficient
        """
        scale_factor = np.log(shear_to / roughness) / (np.log(height / roughness))
        scaled_wspds = wspds * scale_factor

        return scaled_wspds

    @staticmethod
    def _calc_log_law(wspds, heights, return_coeff=False, maximise_data=False) -> (np.array, float):
        """
        Derive the best fit logarithmic law line from a given time-step of speed data at 2 or more elevations

        :param wspds: List of wind speeds [m/s]
        :param heights: List of heights [m above ground]. The position of the height in the list must be the same
            position in the list as its corresponding wind speed value.
        :return: The slope and intercept of the best fit line, as defined above
        :rtype: pandas.Series and float

        METHODOLOGY:
            Derive natural log of elevation data sets
            Derive coefficients of linear best fit along ln(heights)- wspds distribution
            Characterise new distribution of speed values based on linear best fit
            Return the slope and the intercept of this linear best fit.
            The slope and intercept can then be used to find the corresponding roughness coefficient, using the
            equivilant laws:

                                                1)  $U(z) = (v/k)*ln(z/zo)$
            which can be rewritten as:
                                                $U(z) = (v/k)*ln(z) - (v/k)ln(zo)$
                                                where zo = e ** (-c/m) of this line

            Where:

                - U(z) is the wind speed at height z
                - v is the friction velocity at the location
                - k is the Von Karmen constant, taken aa .4
                - z is the height
                - zo is the roughness coefficient

                                                2)  $U2 = (ln(z2/z0)/ln(z1/z0))U1$

            """

        if maximise_data:
            log_heights = np.log(
                pd.Series(heights).drop(wspds[wspds == 0].index.values.astype(int)))  # take log of elevations
            wspds = wspds.drop(wspds[wspds == 0].index.values.astype(int))

        else:
            log_heights = np.log(heights)  # take log of elevations

        coeffs = np.polyfit(log_heights, wspds, deg=1)
        if return_coeff:
            return pd.Series([coeffs[0], coeffs[1]])
        return coeffs[0]

    @staticmethod
    def _calc_power_law(wspds, heights, return_coeff=False, maximise_data=False) -> (np.array, float):
        """
        Derive the best fit power law exponent (as 1/alpha) from a given time-step of speed data at 2 or more elevations

        :param wspds: pandas.Series or list of wind speeds [m/s]
        :param heights: List of heights [m above ground]. The position of the height in the list must be the same
            position in the list as its corresponding wind speed value.
        :return: The shear value (alpha), as the inverse exponent of the best fit power law, based on the form:
            $(v1/v2) = (z1/z2)^(1/alpha)$
        :rtype: pandas.Series and float

        METHODOLOGY:
            Derive natural log of elevation and speed data sets
            Derive coefficients of linear best fit along log-log distribution
            Characterise new distribution of speed values based on linear best fit
            Derive 'alpha' based on gradient of first and last best fit points (function works for 2 or more points)
            Return alpha value

        """
        if maximise_data:
            log_heights = np.log(pd.Series(heights).drop(wspds[wspds == 0].index.values.astype(int)))
            log_wspds = np.log(wspds.drop(wspds[wspds == 0].index.values.astype(int)))

        else:
            log_heights = np.log(heights)  # take log of elevations
            log_wspds = np.log(wspds)  # take log of speeds

        coeffs = np.polyfit(log_heights, log_wspds, deg=1)  # get coefficients of linear best fit to log distribution
        if return_coeff:
            return pd.Series([coeffs[0], np.exp(coeffs[1])])
        return coeffs[0]

    @staticmethod
    def _calc_roughness(slope, intercept):
        return e**(-intercept/slope)

    @staticmethod
    def _by_12x24(wspds, heights, min_speed=3, return_data=False, var_name='Shear'):
        tab_12x24 = dist_12x24(wspds[(wspds > min_speed).all(axis=1)].apply(Shear._calc_power_law, heights=heights,
                                                                            axis=1), return_data=True)[1]
        if return_data:
            return plt.plot_12x24_contours(tab_12x24, label=(var_name, 'mean')), tab_12x24
        return plt.plot_12x24_contours(tab_12x24, label=(var_name, 'mean'))

    @staticmethod
    def scale(wspd,  height, shear_to, alpha=None, roughness=None, calc_method='power_law'):
        """
        Scales wind speeds from one height to another given a value of alpha or roughness coefficient (zo)

        :param wspd: Wind speed time series to apply shear to.
        :type wspd: pandas.Series
        :param height: height of above wspds.
        :type height: float
        :param shear_to: Height to which wspd should be scaled to.
        :type shear_to: float
        :param alpha: Shear exponent to be used when scaling wind speeds.
        :type alpha: Float
        :param roughness: Roughness coefficient to be used when scaling wind speeds.
        :type roughness: float
        :param calc_method: calculation method used to scale the wind speed.
                Using either:   1) 'power_law' :    $(v1/v2) = (z1/z2)^(1/alpha)$
                                2) 'log_law':       $v2 = (ln(z2/z0)/ln(z1/z0))v1$
        :type calc_method: string
        :return: a pandas series of the scaled wind speed
        :return: pandas.Series or float

        **Example Usage**
        ::

           # Scale wind speeds using exponents

           # Specify alpha to use
           alpha_value = .2

           # Specify roughness coefficient to use
           zo = .03

           height = 40
           shear_to = 80

           scaled_by_power_law = bw.Shear.scale(data['Spd40mN'], height, shear_to, alpha=alpha_value)
           scaled_by_log_law = bw.Shear.scale(data['Spd40mN'], height, shear_to, roughness=zo, calc_method='log_law')

        """
        return Shear._scale(wspds=wspd, height=height, shear_to=shear_to, calc_method=calc_method,
                            alpha=alpha, roughness=roughness)

    @staticmethod
    def _scale(wspds, height, shear_to, calc_method='power_law', alpha=None, roughness=None, origin=None):
        """
        Private function for execution of scale()
        """
        if not isinstance(wspds, pd.Series):
            wspds = pd.Series(wspds)

        if calc_method == 'power_law':
            scale_factor = (shear_to / height) ** alpha
            scaled_wspds = wspds * scale_factor

        elif calc_method == 'log_law':
            if origin == 'TimeSeries':
                scaled_wspds = Shear._log_roughness_scale(wspds=wspds, height=height,
                                                          shear_to=shear_to, roughness=roughness)
            else:
                scaled_wspds = wspds.apply(Shear._log_roughness_scale, args=(height, shear_to, roughness))
        else:
            raise ValueError("Please enter a valid calculation method, either 'power_law' or 'log_law'.")

        return scaled_wspds

    @staticmethod
    def _apply(self, wspds, height, shear_to, wdir=None):
        scaled_wspds = pd.Series([])
        result = pd.Series([])

        if self.origin == 'TimeSeries':

            if self.calc_method == 'power_law':
                df = pd.concat([wspds, self.alpha], axis=1).dropna()
                scaled_wspds = Shear._scale(wspds=df.iloc[:, 0], height=height, shear_to=shear_to,
                                            calc_method='power_law', alpha=df.iloc[:, 1])

            else:
                df = pd.concat([wspds, self.roughness], axis=1).dropna()
                scaled_wspds = Shear._scale(wspds=df.iloc[:, 0], height=height, shear_to=shear_to,
                                            calc_method=self.calc_method,
                                            roughness=self._roughness, origin=self.origin)

            result = scaled_wspds.dropna()

        if self.origin == 'TimeOfDay':

            if self.calc_method == 'power_law':
                filled_alpha = Shear._fill_df_12x24(self.alpha)

            else:
                filled_roughness = Shear._fill_df_12x24(self._roughness)
                filled_alpha = filled_roughness

            df_wspds = [[None for y in range(12)] for x in range(24)]
            f = FloatProgress(min=0, max=24 * 12, description='Calculating', bar_style='success')
            display(f)

            for i in range(0, 24):

                for j in range(0, 12):

                    if i == 23:
                        df_wspds[i][j] = wspds[
                            (wspds.index.time >= filled_alpha.index[i]) & (wspds.index.month == j + 1)]
                    else:
                        df_wspds[i][j] = wspds[
                            (wspds.index.time >= filled_alpha.index[i]) & (wspds.index.time < filled_alpha.index[i + 1])
                            & (wspds.index.month == j + 1)]

                    if self.calc_method == 'power_law':
                        df_wspds[i][j] = Shear._scale(df_wspds[i][j], shear_to=shear_to, height=height,
                                                      alpha=filled_alpha.iloc[i, j], calc_method=self.calc_method)

                    else:
                        df_wspds[i][j] = Shear._scale(df_wspds[i][j], shear_to=shear_to, height=height,
                                                      roughness=filled_roughness.iloc[i, j],
                                                      calc_method=self.calc_method)

                    scaled_wspds = pd.concat([scaled_wspds, df_wspds[i][j]], axis=0)
                    f.value += 1

            result = scaled_wspds.sort_index()
            f.close()

        if self.origin == 'BySector':

            # initilise series for later use
            bin_edges = pd.Series([])
            by_sector = pd.Series([])

            if self.calc_method == 'power_law':
                direction_bins = self.alpha
            else:
                direction_bins = self._roughness

            # join wind speeds and directions together in DataFrame
            df = pd.concat([wspds, wdir], axis=1)
            df.columns = ['Unscaled_Wind_Speeds', 'Wind_Direction']

            # get directional bin edges from Shear.by_sector output
            for i in range(self.sectors):
                bin_edges[i] = float(re.findall(r"[-+]?\d*\.\d+|\d+", direction_bins.index[i])[0])
                if i == self.sectors - 1:
                    bin_edges[i + 1] = -float(re.findall(r"[-+]?\d*\.\d+|\d+", direction_bins.index[i])[1])

            for i in range(0, self.sectors):
                if bin_edges[i] > bin_edges[i + 1]:
                    by_sector[i] = df[
                        (df['Wind_Direction'] >= bin_edges[i]) | (df['Wind_Direction'] < bin_edges[i + 1])]

                elif bin_edges[i + 1] == 360:
                    by_sector[i] = df[(df['Wind_Direction'] >= bin_edges[i])]

                else:
                    by_sector[i] = df[
                        (df['Wind_Direction'] >= bin_edges[i]) & (df['Wind_Direction'] < bin_edges[i + 1])]

                by_sector[i].columns = ['Unscaled_Wind_Speeds', 'Wind_Direction']

                if self.calc_method == 'power_law':
                    scaled_wspds[i] = Shear._scale(wspds=by_sector[i]['Unscaled_Wind_Speeds'], height=height,
                                                   shear_to=shear_to,
                                                   calc_method=self.calc_method, alpha=self.alpha[i])

                elif self.calc_method == 'log_law':
                    scaled_wspds[i] = Shear._scale(wspds=by_sector[i]['Unscaled_Wind_Speeds'], height=height,
                                                   shear_to=shear_to,
                                                   calc_method=self.calc_method,
                                                   roughness=self._roughness[i])

                if i == 0:
                    result = scaled_wspds[i]
                else:
                    result = pd.concat([result, scaled_wspds[i]], axis=0)

            result.sort_index(axis='index', inplace=True)

        if self.origin == 'Average':

            if wdir is not None:
                warnings.warn('Warning: Wind direction will not be accounted for when calculating scaled wind speeds.'
                              ' The shear exponents for this object were not calculated by sector. '
                              'Check the origin of the object using ".origin". ')

            if self.calc_method == 'power_law':
                result = Shear._scale(wspds=wspds, height=height, shear_to=shear_to,
                                      calc_method=self.calc_method, alpha=self.alpha)

            elif self.calc_method == 'log_law':
                result = Shear._scale(wspds=wspds, height=height, shear_to=shear_to,
                                      calc_method=self.calc_method, roughness=self._roughness)

        new_name = wspds.name + '_scaled_to_' + str(shear_to) + 'm'
        result.rename(new_name, inplace=True)
        return result

    @staticmethod
    def _fill_df_12x24(data):
        """
        Fills a pandas.DataFrame or Series to be a 12 month x 24 hour pandas.DataFrame by duplicating entries.
        Used for plotting TimeOfDay shear.

        :param data: pandas.DataFrame or Series to be turned into a 12x24 dataframe
        :type data: pandas.Series or pandas.DataFrame.
        :return: 12x24 pandas.DataFrame

        """
        # create copy for later use
        df_copy = data.copy()
        interval = int(24 / len(data))
        # set index for new data frame to deal with less than 24 sectors
        idx = pd.date_range('2017-01-01 00:00', '2017-01-01 23:00', freq='1H')

        # create new dataframe with 24 rows only interval number of unique values
        df = pd.DataFrame(index=pd.DatetimeIndex(idx).time, columns=df_copy.columns)
        df = pd.concat(
            [(df[df_copy.index[0].hour:]), (df[:df_copy.index[0].hour])],
            axis=0)

        for i in range(0, len(df_copy)):
            df.iloc[i * interval:(i + 1) * interval, :] = pd.DataFrame(df_copy.iloc[i, :]).values.T

        df.sort_index(inplace=True)
        df_copy.sort_index(inplace=True)

        if len(df.columns) == 1:
            df_12x24 = pd.DataFrame([[None for y in range(12)] for x in range(24)])
            df_12x24.index = df.index
            for i in range(12):
                df_12x24.iloc[:, i] = df.iloc[:, 0]
            df = df_12x24

        return df

    @staticmethod
    def _data_prep(wspds, heights, min_speed, maximise_data=False):

        if not isinstance(wspds, pd.DataFrame):
            wspds = pd.DataFrame(wspds).T

        if len(wspds.columns) != len(heights):
            Shear._unequal_wspd_heights_error_msg(wspds, heights)

        cvg = None
        if isinstance(wspds.index, pd.DatetimeIndex):
            if maximise_data is False:
                cvg = coverage(wspds[wspds > min_speed].dropna(), period='1AS').sum()[1]
            else:
                _wspds = wspds[wspds > min_speed]
                count = _wspds.count(axis=1)
                count = count[count >= 2]
                count.rename('count', inplace=True)
                cvg = coverage(count, period='1AS').sum()

        wspds = wspds[wspds > min_speed].dropna()

        return wspds, cvg

    @staticmethod
    def _unequal_wspd_heights_error_msg(wspds, heights):
        raise ValueError('An equal number of wind speed data series and heights is required. ' +
                         str(len(wspds.columns)) + ' wind speed(s) and ' +
                         str(len(heights)) + ' height(s) were given.')

    @staticmethod
    def _create_info(self, heights, min_speed, cvg, direction_bin_array=None, segments_per_day=None,
                    segment_start_time=None):

        info = {}
        input_data = {}
        output_data = {}
        input_wind_speeds = {'heights': heights, 'column_names': list(self.wspds.columns.values),
                             'min_spd': min_speed}
        input_data['input_wind_speeds'] = input_wind_speeds
        input_data['calculation_method'] = self.calc_method
        if cvg is not None:
            output_data['concurrent_period'] = float("{:.3f}".format(cvg))

        if self.origin == 'TimeOfDay':
            input_data['segments_per_day'] = int(segments_per_day)
            input_data['segment_start_time'] = int(segment_start_time)

        if self.origin == 'BySector':
            input_wind_dir = {'heights': float((re.findall(r'\d+', str(self.wdir.name))[0])),
                              'column_names': str(self.wdir.name)}
            input_data['input_wind_dir'] = input_wind_dir
            input_data['sectors'] = self.sectors
            input_data['direction_bins'] = direction_bin_array

            if self.calc_method == 'power_law':
                output_data['alpha'] = self.alpha
                output_data['alpha_count'] = self.alpha_count
            if self.calc_method == 'log_law':
                output_data['roughness'] = self.roughness
                output_data['roughness_count'] = self.roughness_count

        if self.calc_method == 'power_law':
            output_data['alpha'] = self.alpha
        if self.calc_method == 'log_law':
            output_data['roughness'] = self.roughness
            
        info['input data'] = input_data
        info['output data'] = output_data

        return info



#     brightwind is a library that provides wind analysts with easy to use tools for working with meteorological data.
#     Copyright (C) 2018 Stephen Holleran, Inder Preet
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Lesser General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
import matplotlib as mpl
from matplotlib.dates import DateFormatter
import calendar
import numpy as np
import pandas as pd
import os
from brightwind.utils import utils
from pandas.plotting import register_matplotlib_converters
from brightwind.utils.utils import _convert_df_to_series
import re
import six

register_matplotlib_converters()

__all__ = ['plot_timeseries',
           'plot_scatter',
           'plot_scatter_wspd',
           'plot_scatter_wdir',
           'plot_shear_by_sector',
           'plot_rose']
#
# try:
#     if 'Gotham Rounded' in \
#             [mpl.font_manager.FontProperties(fname=i).get_name() for i in mpl.font_manager.findSystemFonts()]:
#         mpl.rcParams['font.family'] = 'Gotham Rounded'
# except Exception as ex:
#     raise 'Found exception when checking installed fonts. {}'.format(str(ex))
#
plt.style.use(os.path.join(os.path.dirname(__file__), 'bw.mplstyle'))


class _ColorPalette:

    def __init__(self):
        """
        Color palette to be used for plotting graphs and tables. Colors can be reset by using

        ::
            import brightwind as bw
            bw.analyse.plot.COLOR_PALETTE.primary = '#3366CC'

        Color are called 'primary', secondary', 'tertiary', etc. Lighter and darker shades of primary are called
        'primary_10' for 10% of primary. Gradient goes from 0% (darkest) to 100% (lightest). See
        https://www.w3schools.com/colors/colors_picker.asp for more info.
        """
        self.primary = '#9CC537'        # slightly darker than YellowGreen #9acd32, rgb[156/255, 197/255, 55/255]
        self.secondary = '#2E3743'      # asphalt, rgb[46/255, 55/255, 67/255]
        self.tertiary = '#9B2B2C'       # red'ish, rgb(155, 43, 44)
        self.fourth = '#E57925'         # orange'ish, rgb(229, 121, 37)
        self.fifth = '#F2D869'          # yellow'ish, rgb(242, 216, 105)
        self.sixth = '#AB8D60'
        self.seventh = '#A4D29F'
        self.eighth = '#6E807B'
        self.ninth = '#3D636F'          # blue grey
        self.tenth = '#A49E9D'
        self.eleventh = '#DA9BA6'
        self.primary_10 = '#1F290A'     # darkest green, 10% of primary
        self.primary_35 = '#6C9023'     # dark green, 35% of primary
        self.primary_80 = '#D7EBAD'     # light green, 80% of primary
        self.primary_90 = '#ebf5d6'     # light green, 90% of primary
        self.primary_95 = '#F5FAEA'     # lightest green, 95% of primary
        self.secondary_70 = '#6d737b'   # light asphalt

        _col_map_colors = [self.primary_95,  # lightest primary
                           self.primary,     # primary
                           self.primary_10]  # darkest primary
        self._color_map = self._set_col_map(_col_map_colors)

        self.color_list = [self.primary, self.secondary, self.tertiary, self.fourth, self.fifth, self.sixth,
                           self.seventh, self.eighth, self.ninth, self.tenth, self.eleventh, self.primary_35]

        # set the mpl color cycler to our colors. It has 10 colors
        # mpl.rcParams['axes.prop_cycle']

    @staticmethod
    def _set_col_map(col_map_colors):
        return LinearSegmentedColormap.from_list('color_map', col_map_colors, N=256)

    @property
    def color_map(self):
        return self._color_map

    @color_map.setter
    def color_map(self, col_map_colors):
        self._color_map = self._set_col_map(col_map_colors)


COLOR_PALETTE = _ColorPalette()


def plot_monthly_means(data, coverage=None, ylbl=''):
    fig = plt.figure(figsize=(15, 8))
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
    if len(data.shape) > 1:
        ax.set_prop_cycle(color=COLOR_PALETTE.color_list)
        ax.plot(data, '-o')
        ax.legend(list(data.columns))
    else:
        ax.plot(data, '-o', color=COLOR_PALETTE.primary)
        ax.legend([data.name])
    ax.set_ylabel(ylbl)

    ax.set_xticks(data.index)
    ax.xaxis.set_major_formatter(DateFormatter("%b %Y"))
    fig.autofmt_xdate(rotation=20, ha='center')

    if coverage is not None:
        plot_coverage = True
        if len(coverage.shape) > 1:
            if coverage.shape[1] > 1:
                plot_coverage = False
        if plot_coverage:
            import matplotlib.dates as mdates
            ax.clear()
            ax.plot(data, '-o', color=COLOR_PALETTE.secondary)
            ax2 = ax.twinx()

            for month, coverage in zip(coverage.index, coverage.values):
                ax2.imshow(np.array([[mpl.colors.to_rgb(COLOR_PALETTE.primary)],
                                     [mpl.colors.to_rgb(COLOR_PALETTE.primary_80)]]),
                           interpolation='gaussian', extent=(mdates.date2num(month - pd.Timedelta('10days')),
                                                             mdates.date2num(month + pd.Timedelta('10days')),
                                                             0, coverage), aspect='auto', zorder=1)
                ax2.bar(mdates.date2num(month), coverage, edgecolor=COLOR_PALETTE.secondary, linewidth=0.3,
                        fill=False, zorder=0)

            ax2.set_ylim(0, 1)
            ax.set_ylim(bottom=0)
            ax.set_xlim(data.index[0] - pd.Timedelta('20days'), data.index[-1] + pd.Timedelta('20days'))
            ax.set_zorder(3)
            ax2.yaxis.grid(True)
            ax2.set_axisbelow(True)
            ax.patch.set_visible(False)
            ax2.set_ylabel('Coverage [-]')
            ax2.yaxis.tick_right()
            ax2.yaxis.set_label_position("right")
            plt.close()
            return ax2.get_figure()
    plt.close()
    return ax.get_figure()


def plot_timeseries(data, date_from='', date_to='', y_limits=(None, None)):
    """
    Plot a timeseries of data.

    :param data: Data in the form of a Pandas DataFrame/Series to plot.
    :type data: pd.DataFrame, pd.Series
    :param date_from: Start date used for plotting, if not specified the first timestamp of data is considered. Should
        be in yyyy-mm-dd format
    :type date_from: str
    :param date_to: End date used for plotting, if not specified last timestamp of data is considered. Should
        be in yyyy-mm-dd format
    :type date_to: str
    :param y_limits: y-axis min and max limits. Default is (None, None).
    :type y_limits: tuple, None
    :return: Timeseries plot
    :rtype: matplotlib.figure.Figure

    **Example usage**
    ::
        import brightwind as bw
        data = bw.load_csv(bw.datasets.demo_data)

        # To plot few variables
        bw.plot_timeseries(data[['Spd40mN', 'Spd60mS', 'T2m']])

        # To set a start date
        bw.plot_timeseries(data.Spd40mN, date_from='2017-09-01')

        # To set an end date
        bw.plot_timeseries(data.Spd40mN, date_to='2017-10-01')

        # For specifying a slice
        bw.plot_timeseries(data.Spd40mN, date_from='2017-09-01', date_to='2017-10-01')

        # To set the y-axis minimum to 0
        bw.plot_timeseries(data.Spd40mN, date_from='2017-09-01', date_to='2017-10-01', y_limits=(0, None))

        # To set the y-axis maximum to 25
        bw.plot_timeseries(data.Spd40mN, date_from='2017-09-01', date_to='2017-10-01', y_limits=(0, 25))

    """
    plt.rcParams['figure.figsize'] = (15, 8)    # ** this might be setting the global size which isn't good practice ***
    if isinstance(data, pd.Series):
        data_to_slice = data.copy(deep=False).to_frame()
    else:
        data_to_slice = data.copy()
    sliced_data = utils.slice_data(data_to_slice, date_from, date_to)
    figure = sliced_data.plot(color=COLOR_PALETTE.color_list).get_figure()
    if y_limits is not None:
        figure.axes[0].set_ylim(y_limits)
    plt.close()
    return figure


def _scatter_plot(x, y, predicted_y=None, x_label="Reference", y_label="Target", prediction_marker='-'):
    """
    Plots a scatter plot.
    :param x:
    :param y:
    :param predicted_y: A series of predicted y values after applying the correlation to the x series.
    :param x_label:
    :param y_label:
    :param prediction_marker:
    :return:
    """
    fig, ax = plt.subplots()
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    no_dots = len(x)

    marker_size_max = 216
    marker_size_min = 18
    marker_size = -0.2 * no_dots + marker_size_max  # y=mx+c, m = (216 - 18) / (1000 - 0) i.e. slope changes up to 1000
    marker_size = marker_size_min if marker_size < marker_size_min else marker_size

    max_alpha = 0.7
    min_alpha = 0.3
    alpha = -0.0004 * no_dots + max_alpha  # y=mx+c, m = (0.7 - 0.3) / (1000 - 0) i.e. alpha changes up to 1000 dots
    alpha = min_alpha if alpha < min_alpha else alpha

    ax.scatter(x, y, marker='o', color=COLOR_PALETTE.primary, s=marker_size, alpha=alpha, edgecolors='none')
    fig.set_figwidth(10)
    fig.set_figheight(10.2)
    if predicted_y is not None:
        ax.plot(x, predicted_y, prediction_marker, color=COLOR_PALETTE.secondary)
        ax.legend(['Predicted', 'Original'])
    plt.close()
    return ax.get_figure()


def plot_scatter(x_series, y_series, x_axis_title=None, y_axis_title=None,
                 x_limits=None, y_limits=None):
    """
    Plots a scatter plot of two variable's timeseries.

    :param x_series: The x-axis values or reference variable.
    :type x_series: pd.Series
    :param y_series: The y-axis values or target variable.
    :type y_series: pd.Series
    :param x_axis_title: Title for the x-axis. If None, title will be taken from x_series name.
    :type x_axis_title: str, None
    :param y_axis_title: Title for the y-axis. If None, title will be taken from y_series name.
    :type y_axis_title: str, None
    :param x_limits: x-axis min and max limits.
    :type x_limits: tuple, None
    :param y_limits: y-axis min and max limits.
    :type y_limits: tuple, None
    :return: scatter plot
    :rtype: matplotlib.figure.Figure

    **Example usage**
    ::
        import brightwind as bw
        data = bw.load_csv(bw.datasets.demo_data)

        # To plot two variables against each other
        bw.plot_scatter(data.Spd80mN, data.Spd80mS)

        # To overwrite the default axis titles.
        bw.plot_scatter(data.Dir78mS, data.Dir58mS, x_axis_title='Dir78mS', y_axis_title='Dir58mS')

        # To set the x and y axis limits by using a tuple.
        bw.plot_scatter(data.Dir78mS, data.Dir58mS, x_axis_title='Dir78mS', y_axis_title='Dir58mS',
                        x_limits=(50,300), y_limits=(250,300))

    """
    x_series = _convert_df_to_series(x_series)
    y_series = _convert_df_to_series(y_series)
    if x_axis_title is None:
        x_axis_title = x_series.name
    if y_axis_title is None:
        y_axis_title = y_series.name

    merged_df = pd.concat([x_series, y_series], join='inner', axis=1)
    scat_plot = _scatter_plot(merged_df[x_series.name], merged_df[y_series.name],
                              x_label=x_axis_title, y_label=y_axis_title)

    if x_limits is None:
        x_limits = (round(x_series.min() - 0.5), -(-x_series.max() // 1))
    if y_limits is None:
        y_limits = (round(y_series.min() - 0.5), -(-y_series.max() // 1))
    scat_plot.axes[0].set_xlim(x_limits[0], x_limits[1])
    scat_plot.axes[0].set_ylim(y_limits[0], y_limits[1])
    return scat_plot


def plot_scatter_wdir(x_wdir_series, y_wdir_series, x_axis_title=None, y_axis_title=None,
                      x_limits=(0, 360), y_limits=(0, 360)):
    """
    Plots a scatter plot of two wind direction timeseries and adds a line from 0,0 to 360,360.

    :param x_wdir_series: The x-axis values or reference wind directions.
    :type x_wdir_series: pd.Series
    :param y_wdir_series: The y-axis values or target wind directions.
    :type y_wdir_series: pd.Series
    :param x_axis_title: Title for the x-axis. If None, title will be taken from x_wdir_series name.
    :type x_axis_title: str, None
    :param y_axis_title: Title for the y-axis. If None, title will be taken from y_wdir_series name.
    :type y_axis_title: str, None
    :param x_limits: x-axis min and max limits.
    :type x_limits: tuple
    :param y_limits: y-axis min and max limits.
    :type y_limits: tuple
    :return: scatter plot
    :rtype: matplotlib.figure.Figure

    **Example usage**
    ::
        import brightwind as bw
        data = bw.load_csv(bw.datasets.demo_data)

        #To plot few variables
        bw.plot_scatter_wdir(data.Dir78mS, data.Dir58mS)

        #To overwrite the default axis titles.
        bw.plot_scatter_wdir(data.Dir78mS, data.Dir58mS, x_axis_title='Wind direction at 78m',
                             y_axis_title='Wind direction at 58m')

        #To set the x and y axis limits by using a tuple.
        bw.plot_scatter_wdir(data.Dir78mS, data.Dir58mS, x_axis_title='Reference', y_axis_title='Target',
                             x_limits=(50,300), y_limits=(250,300))

    """
    if x_axis_title is None:
        x_axis_title = x_wdir_series.name + ' [°]'
    if y_axis_title is None:
        y_axis_title = y_wdir_series.name + ' [°]'
    scat_plot = plot_scatter(x_wdir_series, y_wdir_series, x_axis_title=x_axis_title, y_axis_title=y_axis_title,
                             x_limits=x_limits, y_limits=y_limits)
    x = [0, 360]
    y = [0, 360]
    scat_plot.axes[0].plot(x, y, '-', color=COLOR_PALETTE.secondary)
    scat_plot.axes[0].legend(['Reference line', 'Data points'])
    return scat_plot


def plot_scatter_wspd(x_wspd_series, y_wspd_series, x_axis_title=None, y_axis_title=None,
                      x_limits=(0, 30), y_limits=(0, 30)):
    """
    Plots a scatter plot of two wind speed timeseries and adds a reference line from 0,0 to 40,40. This should
    only be used for wind speeds in m/s and not when one of the wind speed series is normalised. Please use the
    basic 'plot_scatter()' function when using normalised wind speeds.

    :param x_wspd_series: The x-axis values or reference wind speeds.
    :type x_wspd_series: pd.Series
    :param y_wspd_series: The y-axis values or target wind speeds.
    :type y_wspd_series: pd.Series
    :param x_axis_title: Title for the x-axis. If None, title will be taken from x_wspd_series name.
    :type x_axis_title: str, None
    :param y_axis_title: Title for the y-axis. If None, title will be taken from y_wspd_series name.
    :type y_axis_title: str, None
    :param x_limits: x-axis min and max limits. Can be set to None to let the code derive the min and max from
                     the x_wspd_series.
    :type x_limits: tuple, None
    :param y_limits: y-axis min and max limits. Can be set to None to let the code derive the min and max from
                     the y_wspd_series.
    :type y_limits: tuple, None
    :return: scatter plot
    :rtype: matplotlib.figure.Figure

    **Example usage**
    ::
        import brightwind as bw
        data = bw.load_csv(bw.datasets.demo_data)

        # To plot two wind speeds against each other
        bw.plot_scatter_wspd(data.Spd80mN, data.Spd80mS)

        # To overwrite the default axis titles.
        bw.plot_scatter_wspd(data.Spd80mN, data.Spd80mS, x_axis_title='Speed at 80m North',
                             y_axis_title='Speed at 80m South')

        # To set the x and y axis limits by using a tuple.
        bw.plot_scatter_wspd(data.Spd80mN, data.Spd80mS, x_axis_title='Speed at 80m North',
                             y_axis_title='Speed at 80m South', x_limits=(0,25), y_limits=(0,25))

    """
    if x_axis_title is None:
        x_axis_title = x_wspd_series.name + ' [m/s]'
    if y_axis_title is None:
        y_axis_title = y_wspd_series.name + ' [m/s]'
    scat_plot = plot_scatter(x_wspd_series, y_wspd_series, x_axis_title=x_axis_title, y_axis_title=y_axis_title,
                             x_limits=x_limits, y_limits=y_limits)
    x = [0, 40]
    y = [0, 40]
    scat_plot.axes[0].plot(x, y, '-', color=COLOR_PALETTE.secondary)
    scat_plot.axes[0].legend(['Reference line', 'Data points'])
    return scat_plot


def plot_freq_distribution(data, max_y_value=None, x_tick_labels=None, x_label=None, y_label=None):
    from matplotlib.ticker import PercentFormatter
    fig = plt.figure(figsize=(15, 8))
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    if isinstance(data.index[0], pd.Interval):
        x_data = [i.mid for i in data.index]
    else:
        x_data = data.index
    ax.set_xticks(x_data)
    ax.set_xlim(x_data[0] - 0.5, x_data[-1] + 0.5)
    if x_tick_labels is not None:
        ax.set_xticklabels(x_tick_labels)
    if max_y_value is None:
        ax.set_ylim(0, data.max() * 1.1)
    else:
        ax.set_ylim(0, max_y_value)
    if y_label[0] == '%':
        ax.yaxis.set_major_formatter(PercentFormatter())
    ax.grid(b=True, axis='y', zorder=0)
    for frequency, ws_bin in zip(data, x_data):
        ax.imshow(np.array([[mpl.colors.to_rgb(COLOR_PALETTE.primary)], [mpl.colors.to_rgb(COLOR_PALETTE.primary_80)]]),
                  interpolation='gaussian', extent=(ws_bin - 0.4, ws_bin + 0.4, 0, frequency), aspect='auto', zorder=3)
        ax.bar(ws_bin, frequency, edgecolor=COLOR_PALETTE.primary_35, linewidth=0.3, fill=False, zorder=5)
    plt.close()
    return ax.get_figure()


def plot_rose(ext_data, plot_label=None):
    """
    Plot a wind rose from data by dist_by_dir_sector
    """
    result = ext_data.copy(deep=False)
    sectors = len(result)
    fig = plt.figure(figsize=(12, 12))
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8], polar=True)
    ax.set_theta_zero_location('N')
    ax.set_theta_direction(-1)
    ax.set_thetagrids(np.arange(0, 360, 360.0/sectors))
    sector_mid_points = []
    widths = []
    for i in result.index:
        angular_pos_start = (np.pi / 180.0) * float(i.split('-')[0])
        angular_pos_end = (np.pi / 180.0) * float(i.split('-')[-1])
        if angular_pos_start < angular_pos_end:
            sector_mid_points.append((angular_pos_start+angular_pos_end)/2.0)
            widths.append(angular_pos_end-angular_pos_start - (np.pi / 180))
        else:
            sector_mid_points.append((np.pi + (angular_pos_start + angular_pos_end)/2.0) % 360)
            widths.append(2*np.pi - angular_pos_start + angular_pos_end - (np.pi / 180))
    max_contour = (ext_data.max() + ext_data.std())
    contour_spacing = max_contour / 10
    num_digits_to_round = 0
    while contour_spacing*(10**num_digits_to_round) <= 1:
        num_digits_to_round += 1
    if 0.5 < contour_spacing < 1:
        contour_spacing = 1
    levels = np.arange(0, max_contour, round(contour_spacing, num_digits_to_round))
    ax.set_rgrids(levels, labels=[str(i) for i in levels], angle=0)
    ax.bar(sector_mid_points, result, width=widths, bottom=0.0, color=COLOR_PALETTE.primary,
           edgecolor=[COLOR_PALETTE.primary_35 for i in range(len(result))], alpha=0.8)
    ax.legend([plot_label])
    plt.close()
    return ax.get_figure()


def plot_rose_with_gradient(freq_table, percent_symbol=True, plot_bins=None, plot_labels=None):
    table = freq_table.copy()
    sectors = len(table.columns)
    table_trans = table.T
    if plot_bins is not None:
        rows_to_sum = []
        intervals = [pd.Interval(plot_bins[i], plot_bins[i + 1], closed=table.index[0].closed)
                     for i in range(len(plot_bins) - 1)]
        bin_assigned = []
        for interval in intervals:
            row_group = []
            for var_bin, pos in zip(table.index, range(len(table.index))):
                if var_bin.overlaps(interval) and not (pos in bin_assigned):
                    bin_assigned.append(pos)
                    row_group.append(pos)
            rows_to_sum.append(row_group)
    else:
        if len(table.index) > 6:
            rows_to_sum = []
            num_rows = len(table.index) // 6
            ctr = 0
            while ctr < len(table.index) - (len(table.index) % 6):
                rows_to_sum.append(list(range(ctr, ctr + num_rows)))
                ctr += num_rows
            rows_to_sum[-1].extend(list(range(len(table.index) - (len(table.index) % 6), len(table.index))))

        else:
            rows_to_sum = [[i] for i in range(len(table.index))]

    table_binned = pd.DataFrame()
    bin_labels = []
    group = 0
    for i in rows_to_sum:
        bin_labels.append(str(table.index[i[0]].left) + ' - ' + str(table.index[i[-1]].right))
        to_concat = table_trans.iloc[:, i].sum(axis=1).rename(group)
        group += 1
        table_binned = pd.concat([table_binned, to_concat], axis=1, sort=True)
    table_binned = table_binned.T
    fig = plt.figure(figsize=(12, 12))
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8], polar=True)
    ax.set_theta_zero_location('N')
    ax.set_theta_direction(-1)
    ax.set_thetagrids(np.arange(0, 360, 360.0 / sectors), zorder=2)

    if percent_symbol:
        symbol = '%'
    else:
        symbol = ' '
    max_contour = max(table.sum(axis=0)) + table.sum(axis=0).std()
    contour_spacing = max_contour / 10
    num_digits_to_round = 0
    while contour_spacing * (10 ** num_digits_to_round) < 1:
        num_digits_to_round += 1
    if 0.5 < contour_spacing < 1:
        contour_spacing = 1
    levels = np.arange(0, max_contour, round(contour_spacing, num_digits_to_round))
    ax.set_rgrids(levels,
                  labels=[str(i) + symbol for i in levels],
                  angle=0, zorder=2)
    ax.set_ylim(0, max(table.sum(axis=0)) + 3.0)
    ax.bar(0, 1, alpha=0)
    norm = mpl.colors.Normalize(vmin=min(table_binned.index), vmax=max(table_binned.index), clip=True)
    mapper = mpl.cm.ScalarMappable(norm=norm, cmap=COLOR_PALETTE.color_map)
    for column in table_binned:
        radial_pos = 0.0
        angular_pos_start = (np.pi / 180.0) * float(column.split('-')[0])
        angular_pos_end = (np.pi / 180.0) * float(column.split('-')[-1])
        # Check for sectors with 0 degrees within the sector
        if angular_pos_end > angular_pos_start:
            angular_width = angular_pos_end - angular_pos_start - (np.pi / 180)  # Leaving 1 degree gap
        else:
            angular_width = 2 * np.pi - angular_pos_start + angular_pos_end - (np.pi / 180)
        for speed_bin, frequency in zip(table_binned.index, table_binned[column]):
            patch = mpl.patches.Rectangle((angular_pos_start, radial_pos), angular_width,
                                          frequency, facecolor=mapper.to_rgba(speed_bin),
                                          edgecolor=COLOR_PALETTE.primary_35,
                                          linewidth=0.3, zorder=3)
            ax.add_patch(patch)
            radial_pos += frequency

    if plot_labels is None:
        plot_labels = [mpl.patches.Patch(color=mapper.to_rgba(table_binned.index[i]), label=bin_labels[i]) for i in
                       range(len(bin_labels))]
    else:
        plot_labels = [mpl.patches.Patch(color=mapper.to_rgba(table_binned.index[i]), label=plot_labels[i]) for i in
                       range(len(plot_labels))]
    ax.legend(handles=plot_labels)
    plt.close()
    return ax.get_figure()


def plot_TI_by_speed(wspd, wspd_std, ti, IEC_class=None):
    """
    Plot turbulence intensity graphs alongside with IEC standards
    :param wspd:
    :param wspd_std:
    :param ti: DataFrame returned from TI.by_speed() in analyse
    :param IEC_class: By default IEC class 2005 is used for custom class pass a DataFrame. Note we have removed
        option to include IEC Class 1999 as no longer appropriate.
        This may need to be placed in a separate function when updated IEC standard is released
    :return: Plots turbulence intensity distribution by wind speed
    """

    # IEC Class 2005

    if IEC_class is None:
        IEC_class = pd.DataFrame(np.zeros([26, 4]), columns=['Windspeed', 'IEC Class A', 'IEC Class B', 'IEC Class C'])
        for n in range(1, 26):
            IEC_class.iloc[n, 0] = n
            IEC_class.iloc[n, 1] = 0.16 * (0.75 + (5.6 / n))
            IEC_class.iloc[n, 2] = 0.14 * (0.75 + (5.6 / n))
            IEC_class.iloc[n, 3] = 0.12 * (0.75 + (5.6 / n))
    common_idxs = wspd.index.intersection(wspd_std.index)
    fig, ax = plt.subplots(figsize=(15, 8))
    ax.scatter(wspd.loc[common_idxs], wspd_std.loc[common_idxs] / wspd.loc[common_idxs],
               color=COLOR_PALETTE.primary, alpha=0.3, marker='.')
    ax.plot(ti.index.values, ti.loc[:, 'Mean_TI'].values, color=COLOR_PALETTE.secondary, label='Mean_TI')
    ax.plot(ti.index.values, ti.loc[:, 'Rep_TI'].values, color=COLOR_PALETTE.primary_35, label='Rep_TI')
    ax.plot(IEC_class.iloc[:, 0], IEC_class.iloc[:, 1], color=COLOR_PALETTE.tertiary, linestyle='dashed')
    ax.plot(IEC_class.iloc[:, 0], IEC_class.iloc[:, 2], color=COLOR_PALETTE.fourth, linestyle='dashed')
    ax.plot(IEC_class.iloc[:, 0], IEC_class.iloc[:, 3], color=COLOR_PALETTE.fifth, linestyle='dashed')
    ax.set_xlim(3, 25)
    ax.set_ylim(0, 0.6)
    ax.set_xticks(np.arange(3, 26, 1))
    ax.set_xlabel('Wind speed [m/s]')
    ax.set_ylabel('Turbulence Intensity')
    ax.grid(True)
    ax.legend()
    plt.close()
    return ax.get_figure()


def plot_TI_by_sector(turbulence, wdir, ti):
    radians = np.radians(utils._get_dir_sector_mid_pts(ti.index))
    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8], polar=True)
    ax.set_theta_zero_location('N')
    ax.set_theta_direction(-1)
    ax.set_thetagrids(utils._get_dir_sector_mid_pts(ti.index))
    ax.plot(np.append(radians, radians[0]), ti.append(ti.iloc[0])['Mean_TI'], color=COLOR_PALETTE.primary, linewidth=4,
            figure=fig)
    maxlevel = ti['Mean_TI'].max() + 0.1
    ax.set_ylim(0, maxlevel)
    ax.scatter(np.radians(wdir), turbulence, color=COLOR_PALETTE.secondary, alpha=0.3, s=1)
    ax.legend(loc=8, framealpha=1)
    plt.close()
    return ax.get_figure()


def plot_shear_by_sector(scale_variable, wind_rose_data, calc_method='power_law'):

    result = wind_rose_data.copy(deep=False)
    radians = np.radians(utils._get_dir_sector_mid_pts(scale_variable.index))
    sectors = len(result)
    fig = plt.figure(figsize=(12, 12),)
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8], polar=True)
    ax.set_theta_zero_location('N')
    ax.set_theta_direction(-1)
    bin_edges = pd.Series([])
    for i in range(sectors):
        bin_edges[i] = float(re.findall(r"[-+]?\d*\.\d+|\d+", wind_rose_data.index[i])[0])
        if i == sectors - 1:
            bin_edges[i + 1] = abs(float(re.findall(r"[-+]?\d*\.\d+|\d+", wind_rose_data.index[i])[1]))
    label = ''
    if calc_method == 'power_law':
        label = 'Mean_Shear'
    if calc_method == 'log_law':
        label = 'Mean_Roughness_Coefficient'

    scale_variable_y = np.append(scale_variable, scale_variable[0])
    plot_x = np.append(radians, radians[0])
    scale_to_fit = max(scale_variable) / max(result/100)
    wind_rose_r = (result/100) * scale_to_fit
    bin_edges = np.array(bin_edges)
    width = pd.Series([])

    for i in range(len(bin_edges) - 1):
        if bin_edges[i + 1] == 0:
            width[i] = 2 * np.pi * (360 - bin_edges[i]) / 360 - (np.pi / 180)
        elif bin_edges[i + 1] > bin_edges[i]:
            width[i] = 2 * np.pi * ((bin_edges[i + 1] - bin_edges[i]) / 360) - (np.pi / 180)
        else:
            width[i] = 2 * np.pi * (((360 + bin_edges[i + 1]) - bin_edges[i]) / 360) - (np.pi / 180)

    ax.bar(radians, wind_rose_r, width=width, color=COLOR_PALETTE.secondary, align='center',
           edgecolor=[COLOR_PALETTE.secondary for i in range(len(result))],
           alpha=0.8, label='Wind_Directional_Frequency')

    maxlevel = (max(scale_variable_y)) + max(scale_variable_y) * .1
    ax.set_thetagrids(radians*180/np.pi)
    ax.plot(plot_x, scale_variable_y, color=COLOR_PALETTE.primary, linewidth=4, label=label)
    ax.set_ylim(0, top=maxlevel)
    ax.legend(loc=8, framealpha=1)

    return ax.get_figure()


def plot_12x24_contours(tab_12x24, label=('Variable', 'mean'), plot=None):
    """
    Get Contour Plot of 12 month x 24 hour matrix of variable
    :param tab_12x24: DataFrame returned from get_12x24() in analyse
    :param label: Label of the colour bar on the plot.
    :return: 12x24 figure
    """
    fig, ax = plt.subplots(figsize=(15, 10))
    x = ax.contourf(tab_12x24.columns, tab_12x24.index, tab_12x24.values, cmap=COLOR_PALETTE.color_map)
    cbar = fig.colorbar(x)
    cbar.ax.set_ylabel(label[1].capitalize() + " of " + label[0])
    ax.set_xlabel('Month of Year')
    ax.set_ylabel('Hour of Day')
    month_names = calendar.month_abbr[1:13]
    ax.set_xticks(tab_12x24.columns)
    ax.set_xticklabels([month_names[i - 1] for i in tab_12x24.columns])
    ax.set_yticks(np.arange(0, 24, 1))
    if plot is None:
        plt.close()
    return ax.get_figure()


def plot_sector_ratio(sec_ratio, wdir, sec_ratio_dist, col_names, boom_dir_1=-1, boom_dir_2=-1):
    """
    Accepts a DataFrame table, along with 2 anemometer names, and one wind vane name and plots the speed ratio
    by sector. Optionally can include anemometer boom directions also.
    :param sec_ratio: Series of sector_ratios
    :type sec_ratio: pandas.Series
    :param wdir: Direction series
    :type wdir: pandas.Series
    :param sec_ratio_dist: DataFrame from SectorRatio.by_sector()
    :type sec_ratio_dist; pandas.Series
    :param boom_dir_1: Boom direction in degrees of speed_col_name_1. Defaults to -1.
    :type boom_dir_1: float
    :param boom_dir_2: Boom direction in degrees of speed_col_name_2. Defaults to -1.
    :type boom_dir_2: float
    :param col_names: A list of strings containing column names of wind speeds, first string is divisor and second is
        dividend
    :type col_names: list(str)
    :returns A speed ratio plot showing average speed ratio by sector and scatter of individual data points.

    """
    radians = np.radians(utils._get_dir_sector_mid_pts(sec_ratio_dist.index))
    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8], polar=True)
    ax.set_theta_zero_location('N')
    ax.set_theta_direction(-1)
    ax.set_thetagrids(utils._get_dir_sector_mid_pts(sec_ratio_dist.index))
    ax.plot(np.append(radians, radians[0]), sec_ratio_dist['Mean_Sector_Ratio'].append(sec_ratio_dist.iloc[0]),
            color=COLOR_PALETTE.primary, linewidth=4)
    # Get max and min levels and set chart axes
    max_level = sec_ratio_dist['Mean_Sector_Ratio'].max() + 0.05
    min_level = sec_ratio_dist['Mean_Sector_Ratio'].min() - 0.1
    ax.set_ylim(min_level, max_level)
    # Add boom dimensions to chart, if required
    width = np.pi / 108
    radii = max_level
    annotate = False
    annotation_text = '* Plot generated using '
    if boom_dir_1 >= 0:
        boom_dir_1_rad = np.radians(boom_dir_1)
        ax.bar(boom_dir_1_rad, radii, width=width, bottom=min_level, color=COLOR_PALETTE.fourth)
        if boom_dir_2 == -1:
            annotation_text += '{} (top mounted) divided by {} ({}° boom)'.format(col_names[1], col_names[0],
                                                                                  boom_dir_1)
            annotate = True
    if boom_dir_2 >= 0:
        boom_dir_2_rad = np.radians(boom_dir_2)
        ax.bar(boom_dir_2_rad, radii, width=width, bottom=min_level, color=COLOR_PALETTE.fifth)
        if boom_dir_1 == -1:
            annotation_text += '{} ({}° boom) divided by {} (top mounted)'.format(col_names[1], boom_dir_2,
                                                                                  col_names[0])
            annotate = True
    if boom_dir_2 >= 0 and boom_dir_1 >= 0:
        annotation_text += '{} ({}° boom) divided by {} ({}° boom)'.format(col_names[1], boom_dir_2,
                                                                           col_names[0], boom_dir_1)
        annotate = True
    if annotate:
        ax.annotate(annotation_text, xy=(0.5, 0.035), xycoords='figure fraction', horizontalalignment='center')
    ax.scatter(np.radians(wdir), sec_ratio, color=COLOR_PALETTE.secondary, alpha=0.3, s=1)
    plt.close()
    return ax.get_figure()


def plot_power_law(avg_alpha, avg_c, wspds, heights, max_plot_height=None, avg_slope=None, avg_intercept=None,
                   plot_both=False):
    if max_plot_height is None:
        max_plot_height = max(heights)

    plot_heights = np.arange(1, max_plot_height+1, 1)
    speeds = avg_c * (plot_heights ** avg_alpha)
    fig, ax = plt.subplots(figsize=(10, 10))
    ax.set_xlabel('Speed [m/s]')
    ax.set_ylabel('Elevation [m]')
    ax.plot(speeds, plot_heights, '-', color=COLOR_PALETTE.primary, label='power_law')
    ax.scatter(wspds, heights, marker='o', color=COLOR_PALETTE.secondary)
    if plot_both is True:
        plot_heights = np.arange(1, max_plot_height+1, 1)
        speeds = avg_slope * np.log(plot_heights) + avg_intercept
        ax.plot(speeds, plot_heights, '-', color=COLOR_PALETTE.secondary, label='log_law')
        ax.scatter(wspds, heights, marker='o', color=COLOR_PALETTE.secondary)
        plt.legend(loc='upper left')

    ax.grid()
    ax.set_xlim(0, max(speeds) + 1)
    ax.set_ylim(0, max(plot_heights) + 10)
    return ax.get_figure()


def plot_log_law(avg_slope, avg_intercept, wspds, heights, max_plot_height=None):
    if max_plot_height is None:
        max_plot_height = max(heights)

    plot_heights = np.arange(1, max_plot_height + 1, 1)
    speeds = avg_slope * np.log(plot_heights) + avg_intercept
    fig, ax = plt.subplots(figsize=(10, 10))
    ax.set_xlabel('Speed [m/s]')
    ax.set_ylabel('Elevation [m]')
    ax.plot(speeds, plot_heights, '-', color=COLOR_PALETTE.primary)
    ax.scatter(wspds, heights, marker='o', color=COLOR_PALETTE.secondary)
    ax.grid()
    ax.set_xlim(0, max(speeds) + 1)
    ax.set_ylim(0, max(plot_heights) + 10)
    return ax.get_figure()


def plot_shear_time_of_day(df, calc_method, plot_type='step'):

    df_copy = df.copy()
    # colours in use
    colors = [(0.6313725490196078, 0.6470588235294118, 0.6705882352941176, 1.0),  # Jan
              (0.1568627450980392, .19215686274509805, 0.6705882352941176, 1.0),  # Feb
              (0.06666666666666667, 0.4196078431372549, 0.6901960784313725, 1.0),  # March
              (0.22745098039215686, 0.7294117647058823, 0.9803921568627451, 1.0),  # April
              (0.2392156862745098, 0.5666666666666667, 0.42745098039215684, 1.0),  # May
              (0.4117647058823529, 0.7137254901960784, 0.16470588235294117, 1.0),  # June
              (0.611764705882353, 0.7725490196078432, 0.21568627450980393, 1.0),  # July
              (0.6823529411764706, 0.403921568627451, 0.1607843137254902, 1.0),  # Aug
              (0.7901960784313726, 0.48627450980392156, 0.1843137254901961, 1.0),  # Sep
              (1, 0.7019607843, .4, 1),  # Oct
              (0, 0, 0, 1.0),  # Nov
              (0.40588235294117647, 0.43137254901960786, 0.4666666666666667, 1.0)]  # Dec

    if len(df.columns) == 1:
        df.columns = ['12 Month Average']
        colors[0] = colors[5]
    if calc_method == 'power_law':
        label = 'Average Shear'

    if calc_method == 'log_law':
        label = 'Roughness Coefficient'

    if plot_type == '12x24':
        df.columns = np.arange(1, 13, 1)
        df.index= np.arange(0, 24, 1)
        df[df.columns[::-1]]
        return plot_12x24_contours(df, label=(label, 'mean'), plot='tod')

    else:
        fig, ax = plt.subplots(figsize=(10, 10))
        ax.set_xlabel('Time of Day')
        ax.set_ylabel(label)
        import matplotlib.dates as mdates

        # create x values for plot
        idx = pd.date_range('2017-01-01 00:00', '2017-01-01 23:00', freq='1H').time

        if plot_type == 'step':
            df = df.shift(+1, axis=0)
            df.iloc[0, :] = df_copy.tail(1).values
            for i in range(0, len(df.columns)):
                ax.step(idx, df.iloc[:, i], label=df.iloc[:, i].name, color=colors[i])

        if plot_type == 'line':
            for i in range(0, len(df.columns)):
                ax.plot(idx, df.iloc[:, i], label=df.iloc[:, i].name, color=colors[i])

        ax.legend(bbox_to_anchor=(1.1, 1.05))
        ax.set_xticks(df.index)
        ax.xaxis.set_minor_formatter(mpl.dates.DateFormatter("%H-%M"))
        _ = plt.xticks(rotation=90)
        return ax.get_figure()


def plot_dist_matrix(matrix, colorbar_label=None, xticklabels=None, yticklabels=None):
    fig, ax = plt.subplots(figsize=(10, 10))
    cm = ax.pcolormesh(matrix, cmap=COLOR_PALETTE.color_map)
    ax.set(xlim=(0, matrix.shape[1]), ylim=(0, matrix.shape[0]))
    ax.set(xticks=np.array(range(0, matrix.shape[1]))+0.5, yticks=np.array(range(0, matrix.shape[0])) + 0.5)
    if xticklabels is not None:
        ax.set_xticklabels(xticklabels)
    if yticklabels is not None:
        ax.set_yticklabels(yticklabels)
    plt.setp(ax.get_xticklabels(), rotation=90, ha="right", rotation_mode="anchor")
    ax.set_xlabel(matrix.columns.names[-1])
    ax.set_ylabel(matrix.index.name)
    cbar = ax.figure.colorbar(cm, ax=ax)
    if colorbar_label is not None:
        cbar.ax.set_ylabel(colorbar_label)
    plt.close()
    return ax.get_figure()


def render_table(data, col_width=3.0, row_height=0.625, font_size=16, header_color=COLOR_PALETTE.primary,
                 row_colors=[COLOR_PALETTE.primary_90, 'w'], edge_color='w', bbox=[0, 0, 1, 1],
                 header_columns=0, show_col_head=1,
                 ax=None, cellLoc='center', padding=0.01, **kwargs):
    if ax is None:
        size = (np.array(data.shape[::-1]) + np.array([0, 1])) * np.array([col_width, row_height])
        fig, ax = plt.subplots(figsize=size)
        ax.axis('off')

    if show_col_head == 1:
        mpl_table = ax.table(cellText=data.values, bbox=bbox, colLabels=data.columns, cellLoc=cellLoc, **kwargs)
    else:
        mpl_table = ax.table(cellText=data.values, bbox=bbox, cellLoc=cellLoc, **kwargs)

    mpl_table.auto_set_font_size(False)
    mpl_table.set_fontsize(font_size)

    for k, cell in six.iteritems(mpl_table._cells):
        cell.set_edgecolor(edge_color)
        if show_col_head == 1:
            if k[0] == 0 or k[1] < header_columns:
                cell.set_text_props(weight='bold', color='w')
                cell.set_facecolor(header_color)
                cell.PAD = padding
            else:
                cell.set_facecolor(row_colors[k[0] % len(row_colors)])
                cell.PAD = padding
        else:
            if k[1] < header_columns:
                cell.set_text_props(weight='bold', color='w')
                cell.set_facecolor(header_color)
                cell.PAD = padding
            else:
                cell.set_facecolor(row_colors[k[0] % len(row_colors)])
                cell.PAD = padding
                # if k[1]==1:
                #   cell.set_width(0.03)
    return ax

# def plot_3d_rose(matrix, colorbar_label=None):
